/*
 * Copyright (c) 1995 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Trevor Blackwell.  Support for use as memcpy() and memmove()
 *	   added by Chris Demetriou.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */
#if defined(MEMCOPY) || defined(MEMMOVE)
# ifdef MEMCOPY
#  define FUNCTION	memcpy
# else
#  define FUNCTION	memmove
# endif
# define SRCREG		$17
# define DSTREG		$16
#else /* !(defined(MEMCOPY) || defined(MEMMOVE)) */
# define FUNCTION	bcopy
# define SRCREG		$16
# define DSTREG		$17
#endif /* !(defined(MEMCOPY) || defined(MEMMOVE)) */

#define	SIZEREG		$18

/*
 * Copy bytes.
 *
 * void bcopy(char *from, char *to, size_t len);
 * char *memcpy(void *to, const void *from, size_t len);
 * char *memmove(void *to, const void *from, size_t len);
 *
 * No matter how invoked, the source and destination registers
 * for calculation.  There's no point in copying them to "working"
 * registers, since the code uses their values "in place," and
 * copying them would be slower.
 */

	.set noreorder
	.set noat

	.align 3

	.globl FUNCTION
	.ent FUNCTION
FUNCTION:
	.frame $30, 0, $26
	.mask 0, 0
	.fmask 0, 0
	.prologue 0

#if defined(MEMCOPY) || defined(MEMMOVE)
	/* set up return value, while we still can */
	mov	SRCREG,$0
#endif

	/* Check for negative length */
	ble	SIZEREG,bcopy_done

	/* Check for overlap */
	subq	DSTREG,SRCREG,$6
	cmpult	$6,SIZEREG,$6
	bne	$6,bcopy_overlap

	/* $19 = end address */
	addq	SRCREG,SIZEREG,$19

	/* Get the first word */
	ldq_u	$3,0(SRCREG)

	/* Do they have the same alignment? */
	xor	SRCREG,DSTREG,$1
	and	$1,7,$1
	and	DSTREG,7,$2
	bne	$1,bcopy_different_alignment

	/* src & dst have same alignment */
	beq	$2,bcopy_all_aligned

	ldq_u	$4,0(DSTREG)
	addq	SIZEREG,$2,SIZEREG
	mskqh	$3,SRCREG,$3
	mskql	$4,SRCREG,$4
	or	$3,$4,$3

	/* Dst is 8-byte aligned */

bcopy_all_aligned:
	/* If less than 8 bytes,skip loop */
	subq	SIZEREG,1,$1
	and	SIZEREG,7,SIZEREG
	bic	$1,7,$1
	beq	$1,bcopy_samealign_lp_end

bcopy_samealign_lp:
	stq_u	$3,0(DSTREG)
	addq	DSTREG,8,DSTREG
	ldq_u	$3,8(SRCREG)
	subq	$1,8,$1
	addq	SRCREG,8,SRCREG
	bne	$1,bcopy_samealign_lp

bcopy_samealign_lp_end:
	/* If we're done, exit */
	bne	SIZEREG,bcopy_small_left
	stq_u	$3,0(DSTREG)
	ret	$31,($26),0

bcopy_small_left:
	mskql	$3,SIZEREG,$5
	ldq_u	$4,0(DSTREG)
	mskqh	$4,SIZEREG,$4
	or	$5,$4,$5
	stq_u	$5,0(DSTREG)
	ret	$31,($26),0

bcopy_different_alignment:
	/*
	 * this is the fun part
	 */
	addq	SRCREG,SIZEREG,$19
	cmpule	SIZEREG,8,$1
	bne	$1,bcopy_da_finish

	beq	$2,bcopy_da_noentry

	/* Do the initial partial word */
	subq	$31,DSTREG,$1
	and	$1,7,$1
	ldq_u	$4,7(SRCREG)
	extql	$3,SRCREG,$3
	extqh	$4,SRCREG,$4
	or	$3,$4,$6
	insql	$6,DSTREG,$6
	ldq_u	$7,0(DSTREG)
	mskql	$7,DSTREG,$7
	or	$6,$7,$6
	stq_u	$6,0(DSTREG)
	addq	SRCREG,$1,SRCREG
	addq	DSTREG,$1,DSTREG
	subq	SIZEREG,$1,SIZEREG
	ldq_u	$3,0(SRCREG)

bcopy_da_noentry:
	subq	SIZEREG,1,$1
	bic	$1,7,$1
	and	SIZEREG,7,SIZEREG
	beq	$1,bcopy_da_finish2

bcopy_da_lp:
	ldq_u	$4,7(SRCREG)
	addq	SRCREG,8,SRCREG
	extql	$3,SRCREG,$5
	extqh	$4,SRCREG,$6
	subq	$1,8,$1
	or	$5,$6,$6
	stq	$6,0(DSTREG)
	addq	DSTREG,8,DSTREG
	beq	$1,bcopy_da_finish1
	ldq_u	$3,7(SRCREG)
	addq	SRCREG,8,SRCREG
	extql	$4,SRCREG,$5
	extqh	$3,SRCREG,$6
	subq	$1,8,$1
	or	$5,$6,$6
	stq	$6,0(DSTREG)
	addq	DSTREG,8,DSTREG
	bne	$1,bcopy_da_lp

bcopy_da_finish2:
	/* Do the last new word */
	mov	$3,$4

bcopy_da_finish1:
	/* Do the last partial word */
	ldq_u	$3,-1($19)
	extql	$4,SRCREG,$4
	extqh	$3,SRCREG,$3
	or	$3,$4,$3
	br	$31,bcopy_samealign_lp_end

bcopy_da_finish:
	/* Do the last word in the next source word */
	ldq_u	$4,-1($19)
	extql	$3,SRCREG,$3
	extqh	$4,SRCREG,$4
	or	$3,$4,$3
	insqh	$3,DSTREG,$4
	insql	$3,DSTREG,$3
	lda	$5,-1($31)
	mskql	$5,SIZEREG,$6
	cmovne	$6,$6,$5
	insqh	$5,DSTREG,$6
	insql	$5,DSTREG,$5
	addq	DSTREG,SIZEREG,$20
	ldq_u	$7,0(DSTREG)
	ldq_u	$8,-1($20)
	bic	$7,$5,$7
	bic	$8,$6,$8
	and	$3,$5,$3
	and	$4,$6,$4
	or	$3,$7,$3
	or	$4,$8,$4
	stq_u	$4,-1($20)
	stq_u	$3,0(DSTREG)
	ret	$31,($26),0

bcopy_overlap:
	/*
	 * Basically equivalent to previous case, only backwards.
	 * Not quite as highly optimized
	 */
	addq	SRCREG,SIZEREG,$19
	addq	DSTREG,SIZEREG,$20

	/* less than 8 bytes - don't worry about overlap */
	cmpule	SIZEREG,8,$1
	bne	$1,bcopy_ov_short

	/* Possibly do a partial first word */
	and	$20,7,$5
	beq	$5,bcopy_ov_nostar$3
	subq	$19,$5,$19
	subq	$20,$5,$20
	ldq_u	$2,0($19)
	subq	SIZEREG,$5,SIZEREG
	ldq_u	$3,7($19)
	ldq	$4,0($20)
	extql	$2,$19,$2
	extqh	$3,$19,$3
	or	$2,$3,$2
	mskqh	$4,$5,$4
	mskql	$2,$5,$2
	or	$2,$4,$2
	stq	$2,0($20)

bcopy_ov_nostar$3:
	bic	SIZEREG,7,$5
	and	SIZEREG,7,SIZEREG
	beq	$5,bcopy_ov_lp_end

bcopy_ov_lp:
	/* This could be more pipelined, but it doesn't seem worth it */
	ldq_u	$1,-8($19)
	subq	$20,8,$20
	ldq_u	$2,-1($19)
	subq	$19,8,$19
	extql	$1,$19,$1
	extqh	$2,$19,$2
	subq	$5,8,$5
	or	$1,$2,$1
	stq	$1,0($20)
	bne	$5,bcopy_ov_lp

bcopy_ov_lp_end:
	beq	SIZEREG,bcopy_done

	ldq_u	$1,0(SRCREG)
	ldq_u	$2,7(SRCREG)
	ldq_u	$3,0(DSTREG)
	extql	$1,SRCREG,$1
	extqh	$2,SRCREG,$2
	or	$1,$2,$1
	insql	$1,DSTREG,$1
	mskql	$3,DSTREG,$3
	or	$3,$1,$3
	stq_u	$3,0(DSTREG)

bcopy_done:
	ret	$31,($26),0

bcopy_ov_short:
	ldq_u	$3,0(SRCREG)
	br	$31,bcopy_da_finish

	.end FUNCTION
